from uiwiz.element import Element

class Dropdown(Element):
    _classes: str = "select select-bordered w-full max-w-xs"

    def __init__(self, items: list[str], placeholder: str = None) -> None:
        super().__init__("select")
        self.classes(Dropdown._classes)
        with self:
            if placeholder:
                Element("option disabled selected", content=placeholder)
            for item in items:
                Element("option", content=item)
        


from uiwiz.element import Element

class Radio(Element):
    _classes: str = "radio"

    def __init__(self, name, value) -> None:
        super().__init__("input")
        self.attributes["name"] = name
        self.attributes["type"] = "radio"
        self.attributes["value"] = value
        self.classes(Radio._classes)


from uiwiz.element import Element
import markdown2


class Markdown(Element):
    def __init__(self, content="") -> None:
        super().__init__(content=content, render_html=False)

    def render_self(self, indent_level=0) -> str:
        return markdown2.markdown(self.content, extras=["fenced-code-blocks"])


from uiwiz.element import Element

class Range(Element):
    _classes: str = "range"

    def __init__(self, min: int, max: int, value: int, name: str, step: int = None) -> None:
        super().__init__("input")
        self.attributes["type"] = "range"
        self.attributes["name"] = name
        self.attributes["min"] = min
        self.attributes["max"] = max
        self.attributes["value"] = value
        if step:
            self.attributes["step"] = step

        self.classes(Range._classes)

    
    @property
    def min(self) -> int:
        return self.attributes["min"]

    @property
    def max(self) -> int:
        return self.attributes["max"]

    @property
    def value(self) -> int:
        return self.attributes["value"]


from typing import Callable
from uiwiz.element import Element

class Input(Element):
    _classes: str = "input input-bordered w-full max-w-xs"

    def __init__(self, placeholder: str = None, name="input", on_change: Callable = None, target: Callable = None, trigger = "input") -> None:
        """Input

        This element is used for input data

        :param placeholder: Text to show before input
        :param name: Used in the attributes name sent back in json
        :param on_change: function or method to call on change
        :param target: Function or id of the element to replace
        :param trigger: keyup, keydown, load and more see htmx docs
        """
        super().__init__("input")
        self.classes(Input._classes)
        self.attributes["name"] = name
        self.attributes["placeholder"] = placeholder
        self.auto_complete = False
        if on_change is not None:
            self.events.append({
                "func": on_change,
                "inputs": None,
                "trigger": trigger,
                "endpoint": None,
                "target": target,
                "swap": "swap"
            })
            
        


import inspect
from pathlib import Path
from typing import Callable, Optional, Union
from fastapi import FastAPI, Response
from fastapi.responses import HTMLResponse
from fastapi.staticfiles import StaticFiles
from fastapi.templating import Jinja2Templates
from starlette.requests import Request
from uiwiz.element import Element, Frame
import functools

class uiwizApp(FastAPI):
    page_routes = {}

    def setup(self):
        self.templates = Jinja2Templates(Path(__file__).parent / 'templates')
        self.add_static_files('/static', Path(__file__).parent / 'static')

    def render(self, html_output: str, request: Request, status_code: int = 200):
        return self.templates.TemplateResponse("default.html",
            {
                'request': request,
                'root_element': [html_output],
            }, status_code, {'Cache-Control': 'no-store', 'X-uiwiz-Content': 'page'})
    
    def render_api(self, html_output: str, status_code: int = 200):
        return HTMLResponse(html_output, status_code, {'Cache-Control': 'no-store', 'X-uiwiz-Content': 'page'})
    
    def route_exists(self, path: str) -> None:
        return path in self.routes
    
    def remove_route(self, path: str) -> None:
        """Remove routes with the given path."""
        self.routes[:] = [r for r in self.routes if getattr(r, 'path', None) != path]

    def add_static_files(self, url_path: str, local_directory: Union[str, Path]) -> None:
        self.mount(url_path, StaticFiles(directory=str(local_directory)))

app = uiwizApp()


class page:
    
    def __init__(self,
                 path: str, *,
                 title: Optional[str] = None,
                 favicon: Optional[str] = None,
                 ) -> None:
        
        """Page

        Creates a new page at the given route.

        :param path: route of the new page (path must start with '/')
        :param title: optional page title
        :param favicon: icon path
        """
        self.path = path
        self.title = title
        self.favicon = favicon

    def __call__(self, func: Callable) -> Callable:
        app.remove_route(self.path)
        parameters_of_decorated_func = list(inspect.signature(func).parameters.keys())


        async def decorated(*dec_args, **dec_kwargs) -> Response:
            frame: Frame = Frame.get_stack()
            frame.app = app
            Element().classes("flex flex-col h-screen")
            request = dec_kwargs['request']
            # NOTE cleaning up the keyword args so the signature is consistent with "func" again
            dec_kwargs = {k: v for k, v in dec_kwargs.items() if k in parameters_of_decorated_func}
            result = func(*dec_args, **dec_kwargs)
            if inspect.isawaitable(result):
                result = await result
            if isinstance(result, Response):  # NOTE if setup returns a response, we don't need to render the page
                return result
            html_output = frame.root_element.render()
            frame.del_stack()
            print(html_output)
            return app.render(html_output, request)

        parameters = [p for p in inspect.signature(func).parameters.values() if p.name != 'client']
        # NOTE adding request as a parameter so we can pass it to the client in the decorated function
        if 'request' not in {p.name for p in parameters}:
            parameters.append(inspect.Parameter('request', inspect.Parameter.POSITIONAL_OR_KEYWORD, annotation=Request))
        decorated.__signature__ = inspect.Signature(parameters)

        app.page_routes[decorated] = self.path

        return app.get(self.path)(decorated)
    
class api:
    def __init__(self, path: str) -> None:
        self.path = path

    def __call__(self, func: Callable) -> Callable:
        parameters_of_decorated_func = list(inspect.signature(func).parameters.keys())

        @functools.wraps(func)
        async def decorated(*dec_args, **dec_kwargs) -> Response:
            frame = Frame.get_stack()
            request = dec_kwargs['request']
            # NOTE cleaning up the keyword args so the signature is consistent with "func" again
            dec_kwargs = {k: v for k, v in dec_kwargs.items() if k in parameters_of_decorated_func}
            result = func(*dec_args, **dec_kwargs)
            if inspect.isawaitable(result):
                result = await result
            if isinstance(result, Response):  # NOTE if setup returns a response, we don't need to render the page
                return result
            html_output = frame.root_element.render()
            frame.del_stack()
            print(html_output)
            return app.render_api(html_output)

        parameters = [p for p in inspect.signature(func).parameters.values() if p.name != 'client']
        if 'request' not in {p.name for p in parameters}:
            parameters.append(inspect.Parameter('request', inspect.Parameter.POSITIONAL_OR_KEYWORD, annotation=Request))
        decorated.__signature__ = inspect.Signature(parameters)

        if not app.route_exists(self.path):
            app.page_routes[decorated] = self.path
        return app.post(self.path)(decorated)
    
Frame.api = api


from uiwiz.element import Element

class Avatar(Element):
    def __init__(self, path: str) -> None:
        super().__init__("div")
        self.classes("avatar")
        with self:
            with Element("div").classes("w-16 rounded-full"):
                img = Element("img")
                img.attributes["src"] = path



from uiwiz.element import Element

class Footer(Element):
    _classes: str = "mt-auto items-center p-4 bg-neutral text-neutral-content"

    def __init__(self) -> None:
        super().__init__(tag="footer")
        self.classes(Footer._classes)


from typing import Callable
from uiwiz.element import Element, Event


class Button(Element):
    _classes: str = "btn btn-neutral place-content-center"

    def __init__(self, title: str) -> None:
        super().__init__(tag="button")
        self.content = title
        self.classes(Button._classes)

    def on_click(self, func: Callable, inputs = None, endpoint=None, target: str = None, swap: str = None) -> "Button":
        self.events.append({
            "func": func,
            "inputs": inputs,
            "trigger": "click",
            "endpoint": endpoint,
            "target": target,
            "swap": swap
        })
        return self



from fastapi import Request
from uiwiz.element import Element
from functools import partial

class Label(Element):
    def __init__(self, text: str = "", for_: Element = None) -> None:
        super().__init__(tag="label")
        self.content = str(text)
        if for_:
            self.attributes["for"] = for_.id

    def set_text(self, text: str) -> "Label":
        self.content = text
        return self
    
    def bind_text_from(self, element: Element, attribute: str, trigger: str = "input delay:20ms"):
        async def bind_Value(request: Request):
            data = await request.json()
            l = Label(data[attribute])
            l.attributes["id"] = self.id

        element.events.append(
            {
                "target": f"#{self.id}",
                "func": bind_Value,
                "trigger": trigger,
            }
        )


from typing import Callable
from uiwiz.element import Element

class TextArea(Element):
    _classes: str = "textarea textarea-bordered"

    def __init__(self, placeholder: str = None, name="input", on_change: Callable = None, target: Callable = None) -> None:
        super().__init__("textarea")
        self.events.append({
            "func": on_change,
            "inputs": None,
            "trigger": "input",
            "endpoint": None,
            "target": target,
            "swap": "swap"
        })
        self.attributes["name"] = name
        self.attributes["placeholder"] = placeholder
        self.classes(TextArea._classes)


import asyncio
from typing import Any, Callable, Optional, TypedDict

class Frame:
    stacks: dict[int, "Frame"] = {}
    api = None

    def __init__(self) -> None:
        self.root_element: Element = None
        self.current_element: Element = None
        self.id = 0
        self.scripts: list[str] = []

    @classmethod
    def get_stack(cls) -> "Frame":
        _id = get_task_id()
        print(_id)
        if _id not in cls.stacks:
            cls.stacks[_id] = Frame()
        return cls.stacks[_id]
    
    @classmethod
    def del_stack(cls):
        del cls.stacks[get_task_id()]
    

def get_task_id() -> int:
    try:
        return id(asyncio.current_task())
    except RuntimeError:
        return 0

class Event(TypedDict):
    func: Callable
    inputs: list[Any]
    trigger: str
    endpoint: Optional[str]
    target: Optional[str]
    swap: Optional[str]
    include: Optional[str]
    vals: Optional[str]


class Element:
    def __init__(self, tag="div", indent_level=2, content="", render_html=True) -> None:
        self.stack = Frame.get_stack()
        self.attributes: dict[str, str] = {}
        self.attributes["id"] = f"a-{self.stack.id}"
        self.stack.id += 1
        self.tag: str = tag

        self.auto_complete: bool = True
        self.events: list[Event] = []
        self.parent_element: Element = None
        self.children: list[Element] = []
        self.script: str = None
        self.render_html: bool = render_html
        self.target: str = "dummyframe" # See default template for hack

        self.content = content
        self.indent = indent_level

        
        if self.stack.root_element is None:
            self.stack.root_element = self
            self.stack.current_element = self
        else:
            self.stack.current_element.children.append(self)
            self.parent_element = self.stack.current_element

    def __enter__(self):
        self.stack.current_element = self
        return self
    
    def __exit__(self, *_):
        self.stack.current_element = self.parent_element
    
    @property
    def id(self):
        return self.attributes["id"]

    def classes(self, input: str):
        self.attributes["class"] = input
        return self

    def render(self, indent_level = 0, render_script_script: bool = True) -> str:
        output = self.render_self()
        if render_script_script:
            output += "<script>"
            for script in self.stack.scripts:
                output += script
            output += "</script>"
        return output
    
    def render_self(self, indent_level = 0) -> str:
        output = ""
        self.before_render()
        if self.render_html:
            output = self.render_attributes(indent_level)
            if self.children:
                output += "\n"
                for child in self.children:
                    output += child.render_self(indent_level + self.indent)
                output += " " * indent_level + f"</{self.tag}>\n"
            else:
                output += self.content
                output += f"</{self.tag}>\n"
        if self.script:
            self.stack.scripts.append(self.script)
        return output

    def render_attributes(self, indent_level) -> str:
        output = " " * indent_level
        output += f'<{self.tag} '
        for key, value in self.attributes.items():
            output += f'{key}="{value}" '
        if not self.auto_complete:
            output += f'autocomplete="off"'
        output = self.render_events(output)
        output += f'>'
        return output
        
    def before_render(self):
        pass

    def render_events(self, output) -> str:
        for event in self.events:
            endpoint = event.get("endpoint")
            if endpoint is None:
                endpoint = "/" + str(hash(frozenset(event.items())))
            
            Frame.api(endpoint)(event["func"])
            target = event.get("target") if event.get("target") is not None else "this"
            swap = event.get("swap") if event.get("swap") is not None else "outerHTML"
            vals = event.get("vals")
            include = event.get("include")
            hx_encoding = event.get("hx-encoding")

            output += f'hx-post="{endpoint}" '
            if event.get("trigger"):
                output += f'hx-trigger="{event["trigger"]}" '

            if isinstance(target, Callable):
                _target = "next #" + str(target())
            else:
                _target = target

            output += f'hx-target="{_target}" '
            output += f'hx-swap="{swap}" '

            if vals:
                output += f"hx-vals='{vals}' "
            if include:
                output += f'hx-include="{include}" '
            if hx_encoding:
                output += f'hx-encoding="{hx_encoding}" '
            else:
                output += "hx-ext='json-enc' "

        return output
    
    def __str__(self) -> str:
        return self.render()
    
    def set_frame(self, frame: Frame):
        self.stack = frame
        for child in self.children:
            child.set_frame(frame)



from uiwiz.element import Element
from uiwiz.elements.label import Label

class Drawer(Element):
    def __init__(self, left=True) -> None:
        super().__init__()
        with self.classes("drawer lg:drawer-open"):
            with Element().classes("drawer-side"):
                with Element("ul").classes("menu p-4 w-80 min-h-full bg-base-200 text-base-content"):
                    Element("li", content="asd")



from uuid import uuid4

from fastapi import Request
from uiwiz.element import Element, Frame

class Tabs(Element):
    _classes: str = "tabs"

    def __init__(self) -> None:
        with Element() as container:
            super().__init__("div")
            self.classes(Tabs._classes)


class Tab(Element):
    _classes: str = "tab tab-bordered"
    _classes_active: str = " tab-active"

    def __init__(self, title: str) -> None:
        super().__init__("a")
        self.classes(Tab._classes)
        self.content = title
        container = self.parent_element.parent_element

        async def handle_tab(request: Request):
            data = await request.json()
            Tab.change_tab(container, data["active_tab"])

        self.events.append({
            "func": handle_tab,
            "inputs": None,
            "trigger": "click",
            "endpoint": None,
            "target": f"#{container.id}",
            "swap": None,
            "vals": f'{{"active_tab": "{title}"}}',
            
        })

    
    def change_tab(container: Element, active_tab: str):
        frame = Frame.get_stack()
        container.set_frame(frame)
        frame.root_element = container

        tabs: TabPanels = Tab.get_tab_panel(container)
        tab: TabPanel
        for tab in tabs.children:
            if tab.tab.content == active_tab:
                tab.tab.classes(Tab._classes + Tab._classes_active)
                tabs.active_tab = tab.tab
                tab.render_html = True
            else:
                tab.tab.classes(Tab._classes)

        pass

    def get_tab_panel(container: Element):
        child: Element
        for child in container.children:
            if isinstance(child, TabPanels):
                return child
            

class TabPanels(Element):
    def __init__(self, tabs: Tabs, active_tab: Tab) -> None:
        super().__init__("div")
        active_tab.attributes["class"] = active_tab.attributes["class"] + Tab._classes_active
        self.tabs = tabs
        self.active_tab = active_tab

    def before_render(self):
        for child in self.children:
            child: TabPanel
            if child.tab != self.active_tab:
                child.render_html = False

    def __exit__(self, *_):
        self.stack.current_element = self.parent_element.parent_element

class TabPanel(Element):
    def __init__(self, tab: Tab) -> None:
        super().__init__("div")
        self.tab = tab



from typing import Callable
from uiwiz.element import Element

class Upload(Element):
    _classes: str = "file-input file-input-bordered w-full max-w-xs"

    def __init__(self, on_upload: Callable, target: Callable, trigger: str = "change") -> None:
        super().__init__("input")
        self.attributes["type"] = "file"
        self.attributes["name"] = "file"
        self.classes(Upload._classes)

        self.events.append({
            "func": on_upload,
            "inputs": None,
            "trigger": trigger,
            "endpoint": None,
            "target": target,
            "swap": "swap",
            "hx-encoding":'multipart/form-data'
        })


from uiwiz.element import Element


class Nav(Element):
    _classes: str = "navbar bg-base-100 bg-error"

    def __init__(self) -> None:
        super().__init__()
        self.classes(Nav._classes)


from typing import Callable
from uiwiz.element import Element

class Form(Element):
    _classes: str = "col"

    def __init__(self) -> None:
        super().__init__("form")
        self.classes(Form._classes)
    
    def on_submit(self, func: Callable, endpoint: str = None):
        self.events.append({
            "func": func,
            "endpoint": endpoint,
            "trigger": "submit",
            "target": "this",
            "swap": "beforeend",
        })
        return self


import pandas as pd
from uiwiz.element import Element
import json

class ToastUIGrid(Element):
    def __init__(self, df: pd.DataFrame, id: str = None) -> None:
        super().__init__("div")
        self.classes("toastgrid")
        cols = []
        rows = []
        if df is not None:
            cols = [{"header": item, "name": item} for item in df.columns.to_list()]
            rows = json.dumps(df.to_dict("records"), default=str)
        self.script = f"""
        var Grid = tui.Grid;
            var instance = new Grid({{
              columnOptions: {{
                minWidth: 180
            }},
            el: document.getElementById('{self.id if id is None else id}'),
            columns: {cols},
            data: {rows},
            usageStatistics: false
        }});

        Grid.applyTheme('striped');
        """


from uiwiz.element import Element

class Divider(Element):
    _classes = "divider"
    _classes_hor = "divider divider-horizontal"

    def __init__(self) -> None:
        super().__init__("div")
        self.classes(Divider._classes)
    
    def horizontal(self) -> "Divider":
        self.classes(Divider._classes_hor)
        return self


from uiwiz.element import Element

class Toast(Element):
    _classes: str = "alert alert-info"

    def __init__(self, message: str) -> None:
        # This toast thing is kind of a hack
        # Use the classes to render the html content
        # and javascript to place it correctly in the DOM.
        # Then some javascript runs on the client to
        # remove it from the DOM
        super().__init__(tag="div")
        self.classes(Toast._classes)
        e = Element("span", content=message)

        output = self.render(render_script_script=False).replace("\n", "")

        self.script = f"""
container = document.getElementById("toast");
$(container).prepend('{output}')""".replace("\n", " ")
        self.render_html = False
        e.render_html = False